"""
FRED (Federal Reserve Economic Data) Data Import

requires API key from FRED
and pip install fredapi
"""
import time
import pandas as pd

try:
    from fredapi import Fred
except Exception:  # except ImportError
    _has_fred = False
else:
    _has_fred = True


def get_fred_data(
    fredkey: str,
    SeriesNameDict: dict = None,
    long=True,
    observation_start=None,
    sleep_seconds: int = 1,
    **kwargs,
):
    """Imports Data from Federal Reserve.
    For simplest results, make sure requested series are all of the same frequency.

    args:
        fredkey (str): an API key from FRED
        SeriesNameDict (dict): pairs of FRED Series IDs and Series Names like: {'SeriesID': 'SeriesName'} or a list of FRED IDs.
            Series id must match Fred IDs, but name can be anything
            if None, several default series are returned
        long (bool): if True, return long style data, else return wide style data with dt index
        observation_start (datetime): passed to Fred get_series
        sleep_seconds (int): seconds to sleep between each series call, reduces failure chance usually
    """
    if not _has_fred:
        raise ImportError("Package fredapi is required")

    fred = Fred(api_key=fredkey)

    if SeriesNameDict is None:
        SeriesNameDict = {
            'T10Y2Y': '10 Year Treasury Constant Maturity Minus 2 Year Treasury Constant Maturity',
            'DGS10': '10 Year Treasury Constant Maturity Rate',
            'DCOILWTICO': 'Crude Oil West Texas Intermediate Cushing Oklahoma',
            'SP500': 'S&P 500',
            'DEXUSEU': 'US Euro Foreign Exchange Rate',
            'DEXCHUS': 'China US Foreign Exchange Rate',
            'DEXCAUS': 'Canadian to US Dollar Exchange Rate Daily',
            'VIXCLS': 'CBOE Volatility Index: VIX',  # this is a more irregular series
            'T10YIE': '10 Year Breakeven Inflation Rate',
            'USEPUINDXD': 'Economic Policy Uncertainty Index for United States',  # also very irregular
        }

    if isinstance(SeriesNameDict, dict):
        series_desired = list(SeriesNameDict.keys())
    else:
        series_desired = list(SeriesNameDict)

    if long:
        fred_timeseries = pd.DataFrame(
            columns=['date', 'value', 'series_id', 'series_name']
        )
    else:
        fred_timeseries = pd.DataFrame()

    for series in series_desired:
        data = fred.get_series(series, observation_start=observation_start)
        try:
            series_name = SeriesNameDict[series]
        except Exception:
            series_name = series

        if long:
            data_df = pd.DataFrame(
                {
                    'date': data.index,
                    'value': data,
                    'series_id': series,
                    'series_name': series_name,
                }
            )
            data_df.reset_index(drop=True, inplace=True)
            fred_timeseries = pd.concat(
                [fred_timeseries, data_df], axis=0, ignore_index=True
            )
        else:
            data.name = series_name
            fred_timeseries = fred_timeseries.merge(
                data, how="outer", left_index=True, right_index=True
            )
        time.sleep(sleep_seconds)

    return fred_timeseries



import numpy as np
import pandas as pd


class hierarchial(object):
    """Create hierarchial series, then reconcile.

    Currently only performs one-level groupings.
    Args:
        grouping_method (str): method to create groups. 'User' requires hier_id input of groupings.
        n_groups (int): number of groups, if above is not 'User'
        reconciliation (str): None, or 'mean' method to combine top and bottom forecasts.
        grouping_ids (dict): dict of series_id: group_id to use if grouping is 'User'
    """

    def __init__(
        self,
        grouping_method: str = 'tile',
        n_groups: int = 5,
        reconciliation: str = 'mean',
        grouping_ids: dict = None,
    ):
        self.grouping_method = str(grouping_method).lower()
        self.n_groups = n_groups
        self.reconciliation = reconciliation
        self.grouping_ids = grouping_ids

        if self.grouping_method == 'user':
            if grouping_ids is None:
                raise ValueError("grouping_ids must be provided.")

    def fit(self, df):
        """Construct and save object info."""
        # construct grouping_ids if not given
        if self.grouping_method != 'user':
            num_hier = df.shape[1] / self.n_groups
            if self.grouping_method == 'dbscan':
                X = df.mean().values.reshape(-1, 1)
                from sklearn.cluster import DBSCAN

                clustering = DBSCAN(eps=0.5, min_samples=2).fit(X)
                grouping_ids = clustering.labels_
            elif self.grouping_method == 'tile':
                grouping_ids = np.tile(
                    np.arange(self.n_groups), int(np.ceil(num_hier))
                )[: df.shape[1]]
            elif self.grouping_method == 'alternating':
                grouping_ids = np.repeat(
                    np.arange(self.n_groups), int(np.ceil(num_hier))
                )[: df.shape[1]]
            elif self.grouping_method == 'kmeans':
                from sklearn.cluster import KMeans

                X = df.mean().values.reshape(-1, 1)
                kmeans = KMeans(n_clusters=self.n_groups, random_state=0).fit(X)
                grouping_ids = kmeans.labels_
            grouping_ids = grouping_ids.astype(str).astype(np.object)
            # z is a deliberate typo to make such an id very rare in source
            grouping_ids = grouping_ids + '_hierarchy_levelz'
            grouping_ids = dict(zip(df.columns.tolist(), grouping_ids))
            self.grouping_ids = grouping_ids
        else:
            # fix missing or extra ids
            grouping_ids = {}
            for x in df.columns:
                if x not in self.grouping_ids.keys():
                    grouping_ids[x] = 'hierarchy_levelz'
                else:
                    grouping_ids[x] = self.grouping_ids[x]
            self.grouping_ids = grouping_ids.copy()

        self.top_ids = set(grouping_ids.values())
        self.bottom_ids = grouping_ids.keys()

        hier = df.abs().groupby(grouping_ids, axis=1).sum()
        self.hier = hier

        if self.reconciliation == 'mean':
            level_sums = pd.DataFrame(hier.sum(axis=0))
            individal_sums = pd.DataFrame(df.abs().sum(axis=0))
            divisors = pd.DataFrame.from_dict(grouping_ids, orient='index')
            divisors.columns = ['group']
            divisors = divisors.merge(level_sums, left_on='group', right_index=True)
            divisors = divisors.merge(individal_sums, left_index=True, right_index=True)
            divisors.columns = ['group', 'divisor', 'value']
            divisors['fraction'] = divisors['value'] / divisors['divisor']
            self.divisors = divisors

        return self

    def transform(self, df):
        """Apply hierarchy to existing data with bottom levels only."""
        try:
            return pd.concat([df, self.hier], axis=1)
        except Exception as e:
            raise ValueError(f"{e} .fit() has not been called.")

    def reconcile(self, df):
        """Apply to forecasted data containing bottom and top levels."""
        if self.reconciliation is None:
            return df[self.bottom_ids]
        elif self.reconciliation == 'mean':
            fore = df
            fracs = pd.DataFrame(
                np.repeat(
                    self.divisors['fraction'].values.reshape(1, -1),
                    fore.shape[0],
                    axis=0,
                )
            )
            fracs.index = fore.index
            fracs.columns = pd.MultiIndex.from_frame(
                self.divisors.reset_index()[['index', 'group']]
            )

            top_level = fore[self.top_ids]
            bottom_up = (
                fore[self.bottom_ids].abs().groupby(self.grouping_ids, axis=1).sum()
            )

            diff = (top_level - bottom_up) / 2

            # gotta love that 'level' option on multiple for easy broadcasting
            test = fracs.multiply(diff, level='group')
            test.columns = self.divisors.index

            result = fore[self.bottom_ids] + test
            return result
        else:
            print("Complete and utter failure.")
            return df


"""
grouping_ids = {
    'CSUSHPISA': 'A',
    'EMVOVERALLEMV': 'A',
    'EXCAUS': 'exchange rates',
    'EXCHUS': 'exchange rates',
    'EXUSEU': 'exchange rates',
    'GS10': 'B',
    'MCOILWTICO': 'C',
    'T10YIEM': 'C',
    'USEPUINDXM': 'C'
    }
test = hierarchial(n_groups=3, grouping_method='dbscan',
                   grouping_ids=None, reconciliation='mean').fit(df)
test_df = test.transform(df)
test.reconcile(test_df)
"""
# how to assign groups
# how to assign blame/reconcile
# multiple group levels
# one additional overall-level



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 15 19:28:57 2023

@author: Colin
"""
import json
import sys
import datetime
import random
import numpy as np
import pandas as pd
from autots.models.base import ModelObject, PredictionObject
from autots.tools.probabilistic import Point_to_Probability
from autots.evaluator.auto_model import RandomTemplate, model_forecast
from autots.models.model_list import diff_window_motif_list, all_result_path
from autots.evaluator.event_forecasting import extract_result_windows
from autots.evaluator.validation import (
    validate_num_validations,
    generate_validation_indices,
)
from autots.models.sklearn import (
    retrieve_regressor,
)  # generate_regressor_params, datepart_model_dict
from autots.tools.shaping import simple_train_test_split
from autots.tools.seasonal import date_part


def create_feature(
    df_train,
    models,
    forecast_length,
    future_regressor_train=None,
    future_regressor_forecast=None,
    datepart_method=None,
):
    result_windows = None
    res = []
    # add last value as a feature
    res.append(
        np.repeat(
            df_train.iloc[-1:,].to_numpy(),
            forecast_length,
            axis=0,
        )
    )
    # add averages
    res.append(
        np.repeat(df_train.mean().to_numpy()[np.newaxis, :], forecast_length, axis=0)
    )
    res.append(
        np.repeat(df_train.std().to_numpy()[np.newaxis, :], forecast_length, axis=0)
    )
    for model in models:
        model_name = model['Model']
        if model_name in diff_window_motif_list:
            model_param_dict = {
                **json.loads(model['ModelParameters']),
                **{"return_result_windows": True},
            }
        else:
            model_param_dict = json.loads(model['ModelParameters'])

        forecasts = model_forecast(
            model_name=model['Model'],
            model_param_dict=model_param_dict,
            model_transform_dict=json.loads(model['TransformationParameters']),
            df_train=df_train,  # .iloc[:, 0:1]
            forecast_length=forecast_length,
            frequency='infer',
            prediction_interval=0.9,
            no_negatives=False,
            future_regressor_train=future_regressor_train,
            future_regressor_forecast=future_regressor_forecast,
            random_seed=321,
            verbose=1,
            n_jobs=4,
            return_model=True,
            fail_on_forecast_nan=True,
        )
        res.extend(
            [forecasts.forecast, forecasts.upper_forecast, forecasts.lower_forecast]
        )
        if model_name == 'Cassandra':
            comps = forecasts.model.return_components()
            ncomps = comps.columns.get_level_values(1)
            for cp in set(ncomps):
                res.append(
                    comps.loc[:, comps.columns.get_level_values(1) == cp]
                    .droplevel(1, 1)
                    .reindex(forecasts.forecast.index)
                )
        elif model_name in all_result_path:
            result_windows = extract_result_windows(forecasts, model_name=model_name)
    res = np.stack(res)
    if result_windows is not None:
        res = np.concatenate([res, result_windows], axis=0)
    # add regressor as a feature
    if future_regressor_forecast is not None:
        r = np.repeat(
            future_regressor_forecast.to_numpy().T[:, :, np.newaxis],
            df_train.shape[1],
            axis=2,
        )
        res = np.concatenate([res, r], axis=0)
    # add timestep as a feature
    res = np.concatenate(
        (
            res,
            np.linspace(
                [0] * df_train.shape[1],
                [res.shape[1] - 1] * df_train.shape[1],
                res.shape[1],
            )[np.newaxis, :, :],
        ),
        axis=0,
    )
    # add time series ID as a feature
    res = np.concatenate(
        (
            res,
            np.linspace(
                [0] * forecast_length,
                [res.shape[2] - 1] * forecast_length,
                res.shape[2],
            ).T[np.newaxis, :, :],
        ),
        axis=0,
    )
    if datepart_method not in [None, "None", "none"]:
        date_part_df = date_part(forecasts.forecast.index, method=datepart_method)
        res = np.concatenate(
            (
                res,
                np.repeat(date_part_df.to_numpy()[np.newaxis, :, :], 21, axis=0).T,
            ),
            axis=0,
        )

    sys.stdout.flush()
    return res


"""
X = []
y = []
for val in range(num_validations + 1):
    df_subset = df_wide_numeric.loc[validation_indexes[val]]
    df_train, df_test = simple_train_test_split(
        df_subset,
        forecast_length=forecast_length,
        min_allowed_train_percent=min_allowed_train_percent,
        verbose=verbose,
    )
    try:
        res = create_feature(df_train, future_regressor, models, forecast_length)
        X.append(res.reshape(res.shape[0], -1))
        y.append(df_test.to_numpy().reshape(-1))
    except Exception as e:
        if val < 1:
            raise e
        else:
            print(f"validation round {y} failed with {repr(e)}")

X = np.concatenate(X, axis=1).T
y = np.concatenate(y, axis=0)


regr = retrieve_regressor(
    regression_model={
        "model": 'XGBRegressor',
        "model_params": {
            "base_score":0.5, "booster":'gbtree', "callbacks":[],
            "colsample_bylevel":0.5050297936099943, "colsample_bynode":1,
            "colsample_bytree":0.616466808436318, "early_stopping_rounds":None,
            "enable_categorical":False, "eval_metric":None, "feature_types":None,
            "gamma":0, "gpu_id":-1, "grow_policy":'depthwise', "importance_type":None,
            "interaction_constraints":'', "learning_rate":0.0026632745639476375,
            "max_bin":256, "max_cat_threshold":64, "max_cat_to_onehot":4,
            "max_delta_step":0, "max_depth":10, "max_leaves":0,
            "min_child_weight":0.010458618463525058,
            "monotone_constraints":'()', "n_estimators":1356,
            "num_parallel_tree":1,
        },
    },
    verbose=verbose,
    verbose_bool=False,
    random_seed=2020,
    n_jobs=1,
    multioutput=False,
)
regr.fit(X, y)


X = []
res = create_feature(df_wide_numeric, future_regressor_future, models, forecast_length)
X.append(res.reshape(res.shape[0], -1))
X = np.concatenate(X, axis=1).T
result = pd.DataFrame(
    regr.predict(X).reshape(forecast_length, df_train.shape[1]),
    index=future_regressor_future.index,
    columns=df_wide_numeric.columns,
)
"""


class MLEnsemble(ModelObject):
    """Combine models using an ML model across validations.

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
    """

    def __init__(
        self,
        name: str = "MLEnsemble",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        holiday_country: str = 'US',
        random_seed: int = 2020,
        verbose: int = 0,
        forecast_length: int = 10,
        regression_type: str = None,
        regression_model=None,
        models=[
            {
                'Model': 'Cassandra',
                'ModelParameters': {},
                "TransformationParameters": {},
            },
            {
                'Model': 'MetricMotif',
                'ModelParameters': {},
                "TransformationParameters": {},
            },
            {
                'Model': 'SeasonalityMotif',
                'ModelParameters': {},
                "TransformationParameters": {},
            },
        ],
        num_validations=2,
        validation_method="backwards",
        min_allowed_train_percent=0.5,
        datepart_method="expanded_binarized",
        models_source: str = 'random',
        **kwargs,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            holiday_country=holiday_country,
            random_seed=random_seed,
            verbose=verbose,
            regression_type=regression_type,
        )
        self.forecast_length = forecast_length
        assert num_validations >= 0, f"window {num_validations} must be >= 0"
        self.num_validations = int(num_validations)
        self.validation_method = validation_method
        self.min_allowed_train_percent = min_allowed_train_percent
        self.datepart_method = datepart_method
        if regression_model is None:
            try:
                import xgboost  # noqa

                regression_model = {
                    "model": 'XGBRegressor',
                    "model_params": {
                        "base_score": 0.5,
                        "booster": 'gbtree',
                        "colsample_bylevel": 0.50502979,
                        "colsample_bynode": 1,
                        "colsample_bytree": 0.6164668,
                        "early_stopping_rounds": None,
                        "enable_categorical": False,
                        "eval_metric": None,
                        "feature_types": None,
                        "gamma": 0,
                        "grow_policy": 'depthwise',
                        "importance_type": None,
                        "interaction_constraints": '',
                        "learning_rate": 0.00266327,
                        "max_bin": 256,
                        "max_cat_threshold": 64,
                        "max_cat_to_onehot": 4,
                        "max_delta_step": 0,
                        "max_depth": 10,
                        "max_leaves": 0,
                        "min_child_weight": 0.0104586,
                        "monotone_constraints": '()',
                        "n_estimators": 1356,
                        "num_parallel_tree": 1,
                    },
                }
            except Exception:
                regression_model = {
                    "model": 'RandomForest',
                    "model_params": {
                        "max_features": 0.16322,
                        "max_leaf_nodes": 10,
                        "n_estimators": 5,
                    },
                }
        self.regression_model = regression_model
        self.models = models
        self.models_source = models_source

        stride_size = round(self.forecast_length / 2)
        stride_size = stride_size if stride_size > 0 else 1
        self.similarity_validation_params = {
            "stride_size": stride_size,
            "distance_metric": "nan_euclidean",
            "include_differenced": True,
            "window_size": 30,
        }
        self.seasonal_validation_params = {
            'window_size': 10,
            'distance_metric': 'mae',
            'datepart_method': 'common_fourier_rw',
        }

    def fit(self, df, future_regressor=None):
        """Train algorithm given data supplied.

        Args:
            df (pandas.DataFrame): Datetime Indexed
            regressor (numpy.Array): additional regressor
        """
        df = self.basic_profile(df)
        self.regressor_train = future_regressor
        if self.regression_type in ["User", "user"] and future_regressor is None:
            raise ValueError("regression_type='User' but no future_regressor passed")
        self.df = df.copy()

        # check how many validations are possible given the length of the data.
        self.num_validations = validate_num_validations(
            self.validation_method,
            self.num_validations,
            df,
            self.forecast_length,
            min_allowed_train_percent=self.min_allowed_train_percent,
            verbose=self.verbose,
        )

        # generate validation indices (so it can fail now, not after all the generations)
        self.validation_indexes = generate_validation_indices(
            self.validation_method,
            self.forecast_length,
            self.num_validations,
            df,
            validation_params=self.similarity_validation_params
            if self.validation_method == "similarity"
            else self.seasonal_validation_params,
            preclean=None,
            verbose=0,
        )
        X = []
        y = []
        for val in range(self.num_validations + 1):
            df_subset = df.loc[self.validation_indexes[val]]
            df_train, df_test = simple_train_test_split(
                df_subset,
                forecast_length=self.forecast_length,
                min_allowed_train_percent=self.min_allowed_train_percent,
                verbose=self.verbose,
            )
            if future_regressor is not None:
                regr_subset_t = future_regressor.reindex(index=df_train.index)
                regr_subset_f = future_regressor.reindex(index=df_test.index)
            else:
                regr_subset_t = None
                regr_subset_f = None
            try:
                self.res = create_feature(
                    df_train,
                    self.models,
                    self.forecast_length,
                    future_regressor_train=regr_subset_t,
                    future_regressor_forecast=regr_subset_f,
                    datepart_method=self.datepart_method,
                )
                X.append(self.res.reshape(self.res.shape[0], -1))
                y.append(df_test.to_numpy().reshape(-1))
            except Exception as e:
                if val < 1:
                    raise e
                else:
                    print(f"validation round {y} failed with {repr(e)}")

        self.X = np.concatenate(X, axis=1).T
        self.y = np.concatenate(y, axis=0)

        self.regr = retrieve_regressor(
            regression_model=self.regression_model,
            verbose=self.verbose,
            verbose_bool=False,
            random_seed=self.random_seed,
            n_jobs=self.n_jobs,
            multioutput=False,
        )
        self.regr.fit(self.X, self.y)
        self.fit_runtime = datetime.datetime.now() - self.startTime
        return self

    def predict(
        self,
        forecast_length: int = None,
        future_regressor=None,
        just_point_forecast=False,
    ):
        """Generates forecast data immediately following dates of index supplied to .fit()

        Args:
            forecast_length (int): Number of periods of data to forecast ahead
            regressor (numpy.Array): additional regressor
            just_point_forecast (bool): If True, return a pandas.DataFrame of just point forecasts

        Returns:
            Either a PredictionObject of forecasts and metadata, or
            if just_point_forecast == True, a dataframe of point forecasts
        """
        predictStartTime = datetime.datetime.now()
        if forecast_length != self.forecast_length:
            raise ValueError("forecast_length must be changed in MLEnsemble init")
        test_index = self.create_forecast_index(forecast_length=self.forecast_length)

        X = []
        res = create_feature(
            self.df,
            self.models,
            self.forecast_length,
            future_regressor_train=self.regressor_train,
            future_regressor_forecast=future_regressor,
            datepart_method=self.datepart_method,
        )
        X.append(res.reshape(res.shape[0], -1))
        X = np.concatenate(X, axis=1).T
        forecast = pd.DataFrame(
            self.regr.predict(X).reshape(self.forecast_length, self.df.shape[1]),
            index=test_index,
            columns=self.df.columns,
        )

        if just_point_forecast:
            return forecast
        else:
            upper_forecast, lower_forecast = Point_to_Probability(
                self.df,
                forecast,
                method='inferred_normal',
                prediction_interval=self.prediction_interval,
            )
            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=forecast_length,
                forecast_index=forecast.index,
                forecast_columns=forecast.columns,
                lower_forecast=lower_forecast,
                forecast=forecast,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )
            return prediction

    def get_new_params(self, method: str = 'random'):
        """Returns dict of new parameters for parameter tuning"""
        if "regressor" in method:
            regression_type_choice = "User"
        else:
            regression_type_choice = None
        # regr_params = generate_regressor_params(model_dict=datepart_model_dict)
        # model3 = random.choice(['Cassandra', 'MetricMotif', 'FBProphet', "SeasonalityMotif", "DatepartRegression"])
        models = RandomTemplate(
            n=3,
            model_list='fast',
            transformer_max_depth=4,
            transformer_list='fast',
        )
        models = models[
            ['Model', 'ModelParameters', 'TransformationParameters']
        ].to_dict(orient='records')

        return {
            "num_validations": random.choices([0, 1, 2], [0.5, 0.3, 0.2])[0],
            "validation_method": random.choices(
                ["backwards", "similarity", "seasonal"], [0.5, 0.3, 0.2]
            )[0],
            "regression_model": None,
            "models": models,
            "datepart_method": random.choices(
                [
                    None,
                    "recurring",
                    "simple",
                    "expanded",
                    "simple_2",
                    "simple_binarized",
                    "expanded_binarized",
                    'common_fourier',
                    'common_fourier_rw',
                ],
                [0.2, 0.2, 0.1, 0.3, 0.3, 0.4, 0.35, 0.45, 0.2],
            )[0],
            "regression_type": regression_type_choice,
            "models_source": 'random',
        }

    def get_params(self):
        """Return dict of current parameters"""
        return {
            "num_validations": self.num_validations,
            "validation_method": self.validation_method,
            "regression_type": self.regression_type,
            "models_source": self.models_source,
            "models": self.models,
            "datepart_method": self.datepart_method,
            "regression_model": self.regression_model,
        }


"""
from flaml import AutoML

# Initialize an AutoML instance
automl = AutoML()
# Specify automl goal and constraint
automl_settings = {
    "time_budget": 120,  # 28800  # in seconds
    "metric": 'mape',
    "task": 'regression',
    "log_file_name": "",
    "early_stop": True,
    "verbose": 1,
    "ensemble": False,
    "n_jobs": 8,
    "free_mem_ratio": 0.2,
    "estimator_list": ['lgbm', 'xgboost', 'xgb_limitdepth', 'rf', 'extra_tree'],
}
automl.fit(X_train=X, y_train=y, **automl_settings)
print(automl.model.estimator)
"""

# run models on latest data



# -*- coding: utf-8 -*-
"""Cassandra testing."""
import unittest
import random
import numpy as np
import pandas as pd
from autots.models.cassandra import Cassandra


class CassandraTest(unittest.TestCase):
    def test_model(self):
        print("Starting Cassandra model tests")
        from autots import load_daily
        import matplotlib.pyplot as plt

        categorical_groups = {
            "wiki_United_States": "country",
            "wiki_Germany": "country",
            "wiki_Jesus": "holiday",
            "wiki_Michael_Jackson": "person",
            "wiki_Easter": "holiday",
            "wiki_Christmas": "holiday",
            "wiki_Chinese_New_Year": "holiday",
            "wiki_Thanksgiving": "holiday",
            "wiki_Elizabeth_II": "person",
            "wiki_William_Shakespeare": "person",
            "wiki_George_Washington": "person",
            "wiki_Cleopatra": "person",
        }
        holiday_countries = {
            "wiki_Elizabeth_II": "uk",
            "wiki_United_States": "us",
            "wiki_Germany": "de",
        }
        df_daily = load_daily(long=False)
        forecast_length = 180
        include_history = True
        df_train = df_daily[:-forecast_length].iloc[:, 1:]
        df_test = df_daily[-forecast_length:].iloc[:, 1:]
        fake_regr = df_daily[:-forecast_length].iloc[:, 0:1]
        fake_regr_fcst = (
            df_daily.iloc[:, 0:1]
            if include_history
            else df_daily[-forecast_length:].iloc[:, 0:1]
        )
        flag_regressor = pd.DataFrame(1, index=fake_regr.index, columns=["flag_test"])
        flag_regressor_fcst = pd.DataFrame(
            1, index=fake_regr_fcst.index, columns=["flag_test"]
        )
        regr_per_series = {
            str(df_train.columns[0]): pd.DataFrame(
                np.random.normal(size=(len(df_train), 1)), index=df_train.index
            )
        }
        regr_per_series_fcst = {
            str(df_train.columns[0]): pd.DataFrame(
                np.random.normal(size=(forecast_length, 1)), index=df_test.index
            )
        }
        constraint = {
            "constraint_method": "quantile",
            "lower_constraint": 0,
            "upper_constraint": None,
            "bounds": True,
        }
        past_impacts = pd.DataFrame(0, index=df_train.index, columns=df_train.columns)
        past_impacts.iloc[-10:, 0] = np.geomspace(1, 10)[0:10] / 100
        past_impacts.iloc[-30:, -1] = np.linspace(1, 10)[0:30] / -100
        future_impacts = pd.DataFrame(0, index=df_test.index, columns=df_test.columns)
        future_impacts.iloc[0:10, 0] = (np.linspace(1, 10)[0:10] + 10) / 100

        c_params = Cassandra().get_new_params()
        c_params = {
            "preprocessing_transformation": {
                "fillna": "ffill",
                "transformations": {0: "Slice", 1: "AlignLastValue"},
                "transformation_params": {
                    0: {"method": 0.5},
                    1: {
                        "rows": 1,
                        "lag": 1,
                        "method": "additive",
                        "strength": 1.0,
                        "first_value_only": False,
                    },
                },
            },
            "scaling": "BaseScaler",
            "seasonalities": ["weekdayofmonth", "common_fourier"],
            "ar_lags": None,
            "ar_interaction_seasonality": None,
            "anomaly_detector_params": {
                "method": "rolling_zscore",
                "transform_dict": {
                    "fillna": None,
                    "transformations": {"0": "ClipOutliers"},
                    "transformation_params": {
                        "0": {"method": "clip", "std_threshold": 6}
                    },
                },
                "method_params": {
                    "distribution": "chi2",
                    "alpha": 0.05,
                    "rolling_periods": 200,
                    "center": False,
                },
                "fillna": "rolling_mean_24",
            },
            "anomaly_intervention": "remove",
            "holiday_detector_params": None,
            "holiday_countries_used": True,
            "multivariate_feature": None,
            "multivariate_transformation": {
                "fillna": "ffill",
                "transformations": {0: "QuantileTransformer", 1: "QuantileTransformer"},
                "transformation_params": {
                    0: {"output_distribution": "uniform", "n_quantiles": 1000},
                    1: {"output_distribution": "uniform", "n_quantiles": 1000},
                },
            },
            "regressor_transformation": {
                "fillna": "rolling_mean_24",
                "transformations": {0: "QuantileTransformer"},
                "transformation_params": {
                    0: {"output_distribution": "uniform", "n_quantiles": 60}
                },
            },
            "regressors_used": True,
            "linear_model": {"model": "lstsq", "lambda": 10, "recency_weighting": None},
            "randomwalk_n": 10,
            "trend_window": 3,
            "trend_standin": "random_normal",
            "trend_anomaly_detector_params": {
                "method": "IQR",
                "transform_dict": {
                    "fillna": "rolling_mean_24",
                    "transformations": {0: "AlignLastValue", 1: "RobustScaler"},
                    "transformation_params": {
                        0: {
                            "rows": 1,
                            "lag": 1,
                            "method": "additive",
                            "strength": 1.0,
                            "first_value_only": False,
                        },
                        1: {},
                    },
                },
                "method_params": {"iqr_threshold": 1.5, "iqr_quantiles": [0.25, 0.75]},
                "fillna": "rolling_mean_24",
            },
            "trend_transformation": {
                "fillna": "ffill",
                "transformations": {0: "SeasonalDifference"},
                "transformation_params": {0: {"lag_1": 7, "method": "LastValue"}},
            },
            "trend_model": {
                "Model": "MetricMotif",
                "ModelParameters": {
                    "window": 30,
                    "point_method": "weighted_mean",
                    "distance_metric": "mqae",
                    "k": 3,
                    "comparison_transformation": {
                        "fillna": "rolling_mean",
                        "transformations": {0: "KalmanSmoothing"},
                        "transformation_params": {
                            0: {
                                "state_transition": [[1]],
                                "process_noise": [[0.064]],
                                "observation_model": [[2]],
                                "observation_noise": 10.0,
                            }
                        },
                    },
                    "combination_transformation": {
                        "fillna": "rolling_mean_24",
                        "transformations": {0: "KalmanSmoothing"},
                        "transformation_params": {
                            0: {
                                "state_transition": [[1]],
                                "process_noise": [[0.246]],
                                "observation_model": [[1]],
                                "observation_noise": 0.5,
                            }
                        },
                    },
                },
            },
            "trend_phi": None,
        }

        mod = Cassandra(
            n_jobs=1,
            **c_params,
            constraint=constraint,
            holiday_countries=holiday_countries,
            max_multicolinearity=0.0001,
        )
        mod.fit(
            df_train,
            categorical_groups=categorical_groups,
            future_regressor=fake_regr,
            regressor_per_series=regr_per_series,
            past_impacts=past_impacts,
            flag_regressors=flag_regressor,
        )
        pred = mod.predict(
            forecast_length=forecast_length,
            include_history=include_history,
            future_regressor=fake_regr_fcst,
            regressor_per_series=regr_per_series_fcst,
            future_impacts=future_impacts,
            flag_regressors=flag_regressor_fcst,
        )
        result = pred.forecast
        series = random.choice(mod.column_names)
        mod.regressors_used
        mod.holiday_countries_used
        start_date = "2019-07-01"
        if False:
            with plt.style.context("seaborn-white"):
                mod.plot_forecast(
                    pred,
                    actuals=df_daily if include_history else df_test,
                    series=series,
                    start_date=start_date,
                )
                plt.show()
                mod.plot_components(
                    pred, series=series, to_origin_space=True, start_date=start_date
                )
                mod.plot_trend(series=series, vline=df_test.index[0], start_date=start_date)
        pred.evaluate(
            df_daily.reindex(result.index)[df_train.columns]
            if include_history
            else df_test[df_train.columns]
        )
        dates = df_daily.index.union(
            mod.create_forecast_index(forecast_length, last_date=df_daily.index[-1])
        )
        regr_ps = {
            "wiki_Germany": regr_per_series_fcst["wiki_Germany"].reindex(
                dates, fill_value=0
            )
        }
        mod.past_impacts_intervention = None
        pred2 = mod.predict(
            forecast_length=forecast_length,
            include_history=True,
            new_df=df_daily[df_train.columns],
            flag_regressors=flag_regressor_fcst.reindex(dates, fill_value=0),
            future_regressor=fake_regr_fcst.reindex(dates, fill_value=0),
            regressor_per_series=regr_ps,
        )
        mod.plot_forecast(pred2, actuals=df_daily, series=series, start_date=start_date)
        mod.return_components()
        print(pred.avg_metrics.round(1))

        self.assertFalse(pred.forecast.isna().all().all())



# -*- coding: utf-8 -*-
"""Anomaly Detector
Created on Mon Jul 18 14:19:55 2022

@author: Colin
"""
import random
import numpy as np
import pandas as pd
from autots.tools.anomaly_utils import (
    anomaly_new_params,
    detect_anomalies,
    limits_to_anomalies,
    anomaly_df_to_holidays,
    holiday_new_params,
    dates_to_holidays,
)
from autots.tools.transform import RandomTransform, GeneralTransformer
from autots.evaluator.auto_model import random_model
from autots.evaluator.auto_model import back_forecast


class AnomalyDetector(object):
    def __init__(
        self,
        output="multivariate",
        method="zscore",
        transform_dict={  # also  suggest DifferencedTransformer
            "transformations": {0: "DatepartRegression"},
            "transformation_params": {
                0: {
                    "datepart_method": "simple_3",
                    "regression_model": {
                        "model": "ElasticNet",
                        "model_params": {},
                    },
                }
            },
        },
        forecast_params=None,
        method_params={},
        eval_period=None,
        n_jobs=1,
    ):
        """Detect anomalies on a historic dataset.
        Note anomaly score patterns vary by method.
        Anomaly flag is standard -1 = anomaly; 1 = regular

        Args:
            output (str): 'multivariate' (each series unique outliers), or 'univariate' (all series together for one outlier flag per timestamp)
            method (str): method choosen, from sklearn, AutoTS, and basic stats. Use `.get_new_params()` to see potential models
            transform_dict (dict): option but helpful, often datepart, differencing, or other standard AutoTS transformer params
            forecast_params (dict): used to backcast and identify 'unforecastable' values, required only for predict_interval method
            method_params (dict): parameters specific to the method, use `.get_new_params()` to see potential models
            eval_periods (int): only use this length tail of data, currently only implemented for forecast_params forecasting if used
            n_jobs (int): multiprocessing jobs, used by some methods

        Methods:
            detect()
            plot()
            get_new_params()
            score_to_anomaly()  # estimate

        Attributes:
            anomalies
            scores
        """
        self.output = output
        self.method = method
        self.transform_dict = transform_dict
        self.forecast_params = forecast_params
        self.method_params = method_params
        self.eval_period = eval_period
        self.n_jobs = n_jobs
        self.anomaly_classifier = None

    def detect(self, df):
        """All will return -1 for anomalies.

        Args:
            df (pd.DataFrame): pandas wide-style data
        Returns:
            pd.DataFrame (classifications, -1 = outlier, 1 = not outlier), pd.DataFrame s(scores)
        """
        self.df = df.copy()
        self.df_anomaly = df.copy()
        if self.transform_dict is not None:
            model = GeneralTransformer(
                **self.transform_dict
            )  # DATEPART, LOG, SMOOTHING, DIFF, CLIP OUTLIERS with high z score
            self.df_anomaly = model.fit_transform(self.df_anomaly)

        if self.forecast_params is not None:
            backcast = back_forecast(
                self.df_anomaly,
                n_splits=self.method_params.get("n_splits", "auto"),
                forecast_length=self.method_params.get("forecast_length", 4),
                frequency="infer",
                eval_period=self.eval_period,
                prediction_interval=self.method_params.get("prediction_interval", 0.9),
                **self.forecast_params,
            )
            # don't difference for prediction_interval
            if self.method not in ["prediction_interval"]:
                if self.eval_period is not None:
                    self.df_anomaly = (
                        self.df_anomaly.tail(self.eval_period) - backcast.forecast
                    )
                else:
                    self.df_anomaly = self.df_anomaly - backcast.forecast

        if not all(self.df_anomaly.columns == df.columns):
            self.df_anomaly.columns = df.columns

        if self.method in ["prediction_interval"]:
            self.anomalies, self.scores = limits_to_anomalies(
                self.df_anomaly,
                output=self.output,
                method_params=self.method_params,
                upper_limit=backcast.upper_forecast,
                lower_limit=backcast.lower_forecast,
            )
        else:
            self.anomalies, self.scores = detect_anomalies(
                self.df_anomaly,
                output=self.output,
                method=self.method,
                transform_dict=self.transform_dict,
                method_params=self.method_params,
                eval_period=self.eval_period,
                n_jobs=self.n_jobs,
            )
        return self.anomalies, self.scores

    def plot(self, series_name=None, title=None, plot_kwargs={}):
        import matplotlib.pyplot as plt

        if series_name is None:
            series_name = random.choice(self.df.columns)
        if title is None:
            title = series_name[0:50] + f" with {self.method} outliers"
        fig, ax = plt.subplots()
        self.df[series_name].plot(ax=ax, title=title, **plot_kwargs)
        if self.output == "univariate":
            i_anom = self.anomalies.index[self.anomalies.iloc[:, 0] == -1]
        else:
            series_anom = self.anomalies[series_name]
            i_anom = series_anom[series_anom == -1].index
        if len(i_anom) > 0:
            ax.scatter(i_anom.tolist(), self.df.loc[i_anom, :][series_name], c="red")

    def fit(self, df):
        return self.detect(df)

    def fit_anomaly_classifier(self):
        """Fit a model to predict if a score is an anomaly."""
        # Using DecisionTree as it should almost handle nonparametric anomalies
        from sklearn.tree import DecisionTreeClassifier

        scores_flat = self.scores.melt(var_name='series', value_name="value")
        categor = pd.Categorical(scores_flat['series'])
        self.score_categories = categor.categories
        scores_flat['series'] = categor
        scores_flat = pd.concat(
            [pd.get_dummies(scores_flat['series']), scores_flat['value']], axis=1
        )
        anomalies_flat = self.anomalies.melt(var_name='series', value_name="value")
        self.anomaly_classifier = DecisionTreeClassifier(max_depth=None).fit(
            scores_flat, anomalies_flat['value']
        )
        # anomaly_classifier.score(scores_flat, anomalies_flat['value'])

    def score_to_anomaly(self, scores):
        """A DecisionTree model, used as models are nonstandard (and nonparametric)."""
        if self.anomaly_classifier is None:
            self.fit_anomaly_classifier()
        scores.index.name = 'date'
        scores_flat = scores.reset_index(drop=False).melt(
            id_vars="date", var_name='series', value_name="value"
        )
        scores_flat['series'] = pd.Categorical(
            scores_flat['series'], categories=self.score_categories
        )
        res = self.anomaly_classifier.predict(
            pd.concat(
                [
                    pd.get_dummies(scores_flat['series'], dtype=float),
                    scores_flat['value'],
                ],
                axis=1,
            )
        )
        res = pd.concat(
            [scores_flat[['date', "series"]], pd.Series(res, name='value')], axis=1
        ).pivot_table(index='date', columns='series', values="value")
        return res[scores.columns]

    @staticmethod
    def get_new_params(method="random"):
        """Generate random new parameter combinations.

        Args:
            method (str): 'fast', 'deep', 'default', or any of the anomaly method names (ie 'IQR') to specify only that method
        """
        forecast_params = None
        method_choice, method_params, transform_dict = anomaly_new_params(method=method)
        if transform_dict == "random":
            transform_dict = RandomTransform(
                transformer_list='fast', transformer_max_depth=2
            )
        if method == "fast":
            preforecast = False
        else:
            preforecast = random.choices([True, False], [0.05, 0.95])[0]

        if preforecast or method_choice == "prediction_interval":
            forecast_params = random_model(
                model_list=['LastValueNaive', 'GLS', 'RRVAR'],
                model_prob=[0.8, 0.1, 0.1],
                transformer_max_depth=5,
                transformer_list="superfast",
                keyword_format=True,
            )
        return {
            "method": method_choice,
            "transform_dict": transform_dict,
            "forecast_params": forecast_params,
            "method_params": method_params,
        }


class HolidayDetector(object):
    def __init__(
        self,
        anomaly_detector_params={},
        threshold=0.8,
        min_occurrences=2,
        splash_threshold=0.65,
        use_dayofmonth_holidays=True,
        use_wkdom_holidays=True,
        use_wkdeom_holidays=True,
        use_lunar_holidays=True,
        use_lunar_weekday=False,
        use_islamic_holidays=True,
        use_hebrew_holidays=True,
        output: str = "multivariate",
        n_jobs: int = 1,
    ):
        """Detect anomalies, then mark as holidays (events, festivals, etc) any that reoccur to a calendar.

        Be aware of timezone, especially combining series from multiple time zones. Dates then may not accurately align.
        Can pick up a holiday on the wrong calendar especially for extended holidays (Christmas) and with short (several years is short here) history.
        Holidays on unusual days or weekdays of month (5th Monday of April) may occur
        No multiyear patterns (election year) are detected - would need lots of history

        Args:
            anomaly_detector_params (dict): anomaly detection params passed to detector class
            threshold (float): percent of date occurrences that must be anomalous (0 - 1)
            splash_threshold (float): None, or % required, avg of nearest 2 neighbors to point
            use* (bool): whether to use these calendars for holiday detection
            output (str): "multivariate" or "univariate", for univariate not all dates_to_holidays styles will work

        Methods:
            detect()
            dates_to_holidays()
            plot()
            get_new_params()
        """
        self.anomaly_detector_params = anomaly_detector_params
        self.threshold = threshold
        self.min_occurrences = min_occurrences
        self.splash_threshold = splash_threshold
        self.use_dayofmonth_holidays = use_dayofmonth_holidays
        self.use_wkdom_holidays = use_wkdom_holidays
        self.use_wkdeom_holidays = use_wkdeom_holidays
        self.use_lunar_holidays = use_lunar_holidays
        self.use_lunar_weekday = use_lunar_weekday
        self.use_islamic_holidays = use_islamic_holidays
        self.use_hebrew_holidays = use_hebrew_holidays
        self.n_jobs = n_jobs
        self.output = output
        self.anomaly_model = AnomalyDetector(
            output=output, **self.anomaly_detector_params, n_jobs=n_jobs
        )

    def detect(self, df):
        """Run holiday detection. Input wide-style pandas time series."""
        self.anomaly_model.detect(df)
        if np.min(self.anomaly_model.anomalies.values) != -1:
            print("No anomalies detected.")
        (
            self.day_holidays,
            self.wkdom_holidays,
            self.wkdeom_holidays,
            self.lunar_holidays,
            self.lunar_weekday,
            self.islamic_holidays,
            self.hebrew_holidays,
        ) = anomaly_df_to_holidays(
            self.anomaly_model.anomalies,
            splash_threshold=self.splash_threshold,
            threshold=self.threshold,
            actuals=df if self.output != "univariate" else None,
            anomaly_scores=self.anomaly_model.scores
            if self.output != "univariate"
            else None,
            use_dayofmonth_holidays=self.use_dayofmonth_holidays,
            use_wkdom_holidays=self.use_wkdom_holidays,
            use_wkdeom_holidays=self.use_wkdeom_holidays,
            use_lunar_holidays=self.use_lunar_holidays,
            use_lunar_weekday=self.use_lunar_weekday,
            use_islamic_holidays=self.use_islamic_holidays,
            use_hebrew_holidays=self.use_hebrew_holidays,
        )
        self.df = df
        self.df_cols = df.columns

    def plot_anomaly(self, kwargs={}):
        self.anomaly_model.plot(**kwargs)

    def plot(
        self,
        series_name=None,
        include_anomalies=True,
        title=None,
        plot_kwargs={},
        series=None,
    ):
        import matplotlib.pyplot as plt

        if series_name is None:
            if series is not None:
                series_name = series
            else:
                series_name = random.choice(self.df.columns)
        if title is None:
            title = (
                series_name[0:50]
                + f" with {self.anomaly_detector_params['method']} holidays"
            )
        fig, ax = plt.subplots()
        self.df[series_name].plot(ax=ax, title=title, **plot_kwargs)
        if include_anomalies:
            # directly copied from above
            if self.anomaly_model.output == "univariate":
                i_anom = self.anomaly_model.anomalies.index[
                    self.anomaly_model.anomalies.iloc[:, 0] == -1
                ]
            else:
                series_anom = self.anomaly_model.anomalies[series_name]
                i_anom = series_anom[series_anom == -1].index
            if len(i_anom) > 0:
                ax.scatter(
                    i_anom.tolist(), self.df.loc[i_anom, :][series_name], c="red"
                )
        # now the actual holidays
        i_anom = self.dates_to_holidays(self.df.index, style="series_flag")[series_name]
        i_anom = i_anom.index[i_anom == 1]
        if len(i_anom) > 0:
            ax.scatter(i_anom.tolist(), self.df.loc[i_anom, :][series_name], c="green")

    def dates_to_holidays(self, dates, style="flag", holiday_impacts=False):
        """Populate date information for a given pd.DatetimeIndex.

        Args:
            dates (pd.DatetimeIndex): list of dates
            day_holidays (pd.DataFrame): list of month/day holidays. Pass None if not available
            style (str): option for how to return information
                "long" - return date, name, series for all holidays in a long style dataframe
                "impact" - returns dates, series with values of sum of impacts (if given) or joined string of holiday names
                'flag' - return dates, holidays flag, (is not 0-1 but rather sum of input series impacted for that holiday and day)
                'prophet' - return format required for prophet. Will need to be filtered on `series` for multivariate case
                'series_flag' - dates, series 0/1 for if holiday occurred in any calendar
            holiday_impacts (dict): a dict passed to .replace contaning values for holiday_names, or str 'value' or 'anomaly_score'
        """
        return dates_to_holidays(
            dates,
            self.df_cols,
            style=style,
            holiday_impacts=holiday_impacts,
            day_holidays=self.day_holidays,
            wkdom_holidays=self.wkdom_holidays,
            wkdeom_holidays=self.wkdeom_holidays,
            lunar_holidays=self.lunar_holidays,
            lunar_weekday=self.lunar_weekday,
            islamic_holidays=self.islamic_holidays,
            hebrew_holidays=self.hebrew_holidays,
        )

    def fit(self, df):
        return self.detect(df)

    @staticmethod
    def get_new_params(method="random"):
        holiday_params = holiday_new_params(method=method)
        holiday_params['anomaly_detector_params'] = AnomalyDetector.get_new_params(
            method=method
        )
        return holiday_params



"""Phases of the moon.
Modified from https://stackoverflow.com/a/2531541/9492254
by keturn and earlier from John Walker
"""

from math import sin, cos, floor, sqrt, pi
import bisect
import numpy as np
import pandas as pd


def moon_phase(
    datetime_index,
    epsilon=1e-6,
    epoch=2444237.905,
    ecliptic_longitude_epoch=278.833540,
    ecliptic_longitude_perigee=282.596403,
    eccentricity=0.016718,
    moon_mean_longitude_epoch=64.975464,
    moon_mean_perigee_epoch=349.383063,
):
    """Numpy version. Takes a pd.DatetimeIndex and returns moon phase (%illuminated).
    Epoch can be adjust slightly (0.5 = half day) to adjust for time zones. This is for US. epoch=2444238.5 for Asia generally.
    """
    # set time to Noon if not otherwise given, as midnight is confusingly close to previous day
    if np.sum(datetime_index.hour) == 0:
        datetime_index = datetime_index + pd.Timedelta(hours=12)
    days = datetime_index.to_julian_date() - epoch

    # Mean anomaly of the Sun
    a = (360 / 365.2422) * days
    N = a - 360.0 * np.floor(a / 360.0)
    N = N + ecliptic_longitude_epoch - ecliptic_longitude_perigee
    # Convert from perigee coordinates to epoch 1980
    M = a - 360.0 * np.floor(N / 360.0)

    m = torad(M)
    e = m.copy()
    while 1:
        delta = e - eccentricity * np.sin(e) - m
        e = e - delta / (1.0 - eccentricity * np.cos(e))
        if abs(delta).max() <= epsilon:
            break

    Ec = sqrt((1 + eccentricity) / (1 - eccentricity)) * np.tan(e / 2.0)
    # True anomaly
    Ec = 2 * todeg(np.arctan(Ec))
    # Suns's geometric ecliptic longuitude
    a = Ec + ecliptic_longitude_perigee
    lambda_sun = a - 360.0 * np.floor(a / 360.0)

    # Calculation of the Moon's position

    # Moon's mean longitude
    a = 13.1763966 * days + moon_mean_longitude_epoch
    moon_longitude = a - 360.0 * np.floor(a / 360.0)

    # Moon's mean anomaly
    a = moon_longitude - 0.1114041 * days - moon_mean_perigee_epoch
    MM = a - 360.0 * np.floor(a / 360.0)

    # Moon's ascending node mean longitude
    # MN = fixangle(c.node_mean_longitude_epoch - 0.0529539 * day)

    evection = 1.2739 * np.sin(torad(2 * (moon_longitude - lambda_sun) - MM))

    # Annual equation
    annual_eq = 0.1858 * np.sin(torad(M))

    # Correction term
    A3 = 0.37 * np.sin(torad(M))

    MmP = MM + evection - annual_eq - A3

    # Correction for the equation of the centre
    mEc = 6.2886 * np.sin(torad(MmP))

    # Another correction term
    A4 = 0.214 * np.sin(torad(2 * MmP))

    # Corrected longitude
    lP = moon_longitude + evection + mEc - annual_eq + A4

    # Variation
    variation = 0.6583 * np.sin(torad(2 * (lP - lambda_sun)))

    # True longitude
    lPP = lP + variation

    # Calculation of the phase of the Moon

    # Age of the Moon, in degrees
    moon_age = lPP - lambda_sun

    # Phase of the Moon
    moon_phase = (1 - np.cos(torad(moon_age))) / 2.0
    return moon_phase
    # return pd.Series(moon_phase, index=datetime_index)


def moon_phase_df(datetime_index, epoch=2444237.905):
    """Convert pandas DatetimeIndex to moon phases. Note timezone and hour can matter slightly.
    Epoch can be adjust slightly (0.5 = half day) to adjust for time zones.
    2444237.905 is for US Central. epoch=2444238.5 for Asia generally.
    """
    moon = pd.Series(moon_phase(datetime_index, epoch=epoch), index=datetime_index)
    full_moon = ((moon > moon.shift(1)) & (moon > moon.shift(-1))).astype(int)
    new_moon = ((moon < moon.shift(1)) & (moon < moon.shift(-1))).astype(int)
    # account for end (shift) being new_moon
    if new_moon.tail(29).sum() == 0:
        new_moon.iloc[-1] = 1
    if full_moon.tail(29).sum() == 0:
        full_moon.iloc[-1] = 1
    moon_df = pd.concat([moon, full_moon, new_moon], axis=1)
    moon_df.columns = ['phase', 'full_moon', 'new_moon']
    return moon_df


# Little mathematical functions
def fixangle(a):
    return a - 360.0 * floor(a / 360.0)


def torad(d):
    return d * pi / 180.0


def todeg(r):
    return r * 180.0 / pi


def dsin(d):
    return sin(torad(d))


def dcos(d):
    return cos(torad(d))


def kepler(m, ecc=0.016718):
    """Solve the equation of Kepler."""

    epsilon = 1e-6

    m = torad(m)
    e = m
    while 1:
        delta = e - ecc * sin(e) - m
        e = e - delta / (1.0 - ecc * cos(e))

        if abs(delta) <= epsilon:
            break

    return e


def phase_string(
    p, precision=0.05, new=0.0, first=0.25, full=0.4, last=0.75, nextnew=1.0
):
    phase_strings = (
        (new + precision, "new"),
        (first - precision, "waxing crescent"),
        (first + precision, "first quarter"),
        (full - precision, "waxing gibbous"),
        (full + precision, "full"),
        (last - precision, "waning gibbous"),
        (last + precision, "last quarter"),
        (nextnew - precision, "waning crescent"),
        (nextnew + precision, "new"),
    )

    i = bisect.bisect([a[0] for a in phase_strings], p)

    return phase_strings[i][1]



"""Lists of models grouped by aspects."""
all_models = [
    'ConstantNaive',
    'LastValueNaive',
    'AverageValueNaive',
    'GLS',
    'GLM',
    'ETS',
    'ARIMA',
    'FBProphet',
    'RollingRegression',
    'GluonTS',
    'SeasonalNaive',
    'UnobservedComponents',
    # 'VARMAX',
    'VECM',
    'DynamicFactor',
    'MotifSimulation',
    'WindowRegression',
    'VAR',
    'DatepartRegression',
    "UnivariateRegression",
    # "Greykite",
    'UnivariateMotif',
    'MultivariateMotif',
    'NVAR',
    'MultivariateRegression',
    'SectionalMotif',
    'Theta',
    'ARDL',
    'NeuralProphet',
    'DynamicFactorMQ',
    'PytorchForecasting',
    'ARCH',
    'RRVAR',
    'MAR',
    'TMF',
    'LATC',
    'KalmanStateSpace',
    'MetricMotif',
    'Cassandra',
    'SeasonalityMotif',
    'MLEnsemble',
    'PreprocessingRegression',
]
all_pragmatic = list((set(all_models) - set(['MLEnsemble', 'VARMAX', 'Greykite'])))
# downweight slower models
default = {
    'ConstantNaive': 1,
    'LastValueNaive': 1,
    'AverageValueNaive': 1,
    'GLS': 1,
    'SeasonalNaive': 1,
    'GLM': 1,
    'ETS': 1,
    'FBProphet': 0.5,
    'GluonTS': 0.5,
    'UnobservedComponents': 1,
    'VAR': 1,
    'VECM': 1,
    'ARIMA': 0.4,
    'WindowRegression': 0.5,
    'DatepartRegression': 1,
    'UnivariateRegression': 0.3,
    'MultivariateRegression': 0.4,
    'UnivariateMotif': 1,
    'MultivariateMotif': 1,
    'SectionalMotif': 1,
    'NVAR': 1,
    'Theta': 1,
    'ARDL': 1,
    'ARCH': 1,
    'MetricMotif': 1,
    # 'SeasonalityMotif': 1,
}
# fastest models at any scale
superfast = [
    'ConstantNaive',
    'LastValueNaive',
    'AverageValueNaive',
    'GLS',
    'SeasonalNaive',
    # 'MetricMotif',
    'SeasonalityMotif',
]
# relatively fast
fast = {
    'ConstantNaive': 1,
    'LastValueNaive': 1.5,
    'AverageValueNaive': 1,
    'GLS': 1,
    'SeasonalNaive': 1,
    'GLM': 1,
    'ETS': 1,
    'VAR': 0.8,
    'VECM': 1,
    'WindowRegression': 0.5,  # this gets slow with Transformer, KerasRNN
    'DatepartRegression': 0.8,
    'UnivariateMotif': 1,
    'MultivariateMotif': 0.8,
    'SectionalMotif': 1,
    'NVAR': 1,
    'MAR': 1,
    'RRVAR': 1,
    'KalmanStateSpace': 1,
    'MetricMotif': 1,
    'Cassandra': 1,
    'SeasonalityMotif': 1,
}
# models that can scale well if many CPU cores are available
parallel = {
    'ETS': 1,
    'FBProphet': 0.8,
    'ARIMA': 0.2,  # slow
    'GLM': 1,
    'UnobservedComponents': 1,
    # "Greykite": 0.3,
    'UnivariateMotif': 1,
    'MultivariateMotif': 1,
    'Theta': 1,
    'ARDL': 1,
    'ARCH': 1,
}
# models that should be fast given many CPU cores
fast_parallel = {**parallel, **fast}
fast_parallel_no_arima = {
    i: fast_parallel[i]
    for i in fast_parallel
    if i not in ['ARIMA', 'NVAR', "UnobservedComponents"]
}
# so this opiniated and not fully updated always
best = list(
    set(
        list(fast_parallel_no_arima.keys())
        + ['MultivariateRegression', 'GluonTS', 'PytorchForecasting']
    )
)

# models that are explicitly not production ready
experimental = [
    'MotifSimulation',
    'TensorflowSTS',
    'ComponentAnalysis',
    'TFPRegression',
]
# models that perform slowly at scale
slow = list((set(all_models) - set(fast.keys())) - set(experimental))
# use GPU
gpu = ['GluonTS', 'WindowRegression', 'PytorchForecasting']
# models with model-based upper/lower forecasts
probabilistic = [
    'ARIMA',
    'GluonTS',
    'FBProphet',
    'AverageValueNaive',
    # 'VARMAX',  # yes but so slow
    'DynamicFactor',
    'VAR',
    'UnivariateMotif',
    "MultivariateMotif",
    'SectionalMotif',
    'NVAR',
    'Theta',
    'ARDL',
    'UnobservedComponents',
    'DynamicFactorMQ',
    'PytorchForecasting',
    # 'MultivariateRegression',
    'ARCH',
    'KalmanStateSpace',
    'MetricMotif',
    'Cassandra',
    'SeasonalityMotif',
]
# models that use the shared information of multiple series to improve accuracy
multivariate = [
    'VECM',
    'DynamicFactor',
    'GluonTS',
    # 'VARMAX',  # yes but so slow
    'RollingRegression',
    'WindowRegression',
    'VAR',
    "MultivariateMotif",
    'NVAR',
    'MultivariateRegression',
    'SectionalMotif',
    'DynamicFactorMQ',
    'PytorchForecasting',
    'RRVAR',
    'MAR',
    'TMF',
    'LATC',
    'Cassandra',  # depends
]
univariate = list((set(all_models) - set(multivariate)) - set(experimental))
# USED IN AUTO_MODEL, models with no parameters
no_params = ['LastValueNaive', 'GLS']
# USED IN AUTO_MODEL, ONLY MODELS WHICH CAN ACCEPT RANDOM MIXING OF PARAMS
recombination_approved = [
    'SeasonalNaive',
    'MotifSimulation',
    "ETS",
    'DynamicFactor',
    'VECM',
    'VARMAX',
    'GLM',
    'ARIMA',
    'FBProphet',
    'GluonTS',
    'RollingRegression',
    'VAR',
    # 'WindowRegression',
    'TensorflowSTS',
    'TFPRegression',
    'UnivariateRegression',
    "Greykite",
    'UnivariateMotif',
    "MultivariateMotif",
    'NVAR',
    'MultivariateRegression',
    'SectionalMotif',
    'Theta',
    'ARDL',
    'NeuralProphet',
    'DynamicFactorMQ',
    'PytorchForecasting',
    'ARCH',
    'RRVAR',
    'MAR',
    'TMF',
    'LATC',
    # 'KalmanStateSpace', # matrix sizes must match
    'MetricMotif',
    'Cassandra',
    'SeasonalityMotif',
    'PreprocessingRegression',
]
# USED IN AUTO_MODEL for models that don't share information among series
no_shared = [
    'ConstantNaive',
    'LastValueNaive',
    'AverageValueNaive',
    'GLM',
    'ETS',
    'ARIMA',
    'FBProphet',
    'SeasonalNaive',
    'UnobservedComponents',
    'TensorflowSTS',
    "GLS",
    "UnivariateRegression",
    "Greykite",
    'UnivariateMotif',
    'Theta',
    'ARDL',
    'NeuralProphet',
    'ARCH',
    'KalmanStateSpace',
    'MetricMotif',
    'SeasonalityMotif',
]
# allow the use of a regressor, need to accept "User" (fail if not given), have 'regressor' param method
regressor = [
    'GLM',
    'ARIMA',
    'FBProphet',
    'RollingRegression',
    'UnobservedComponents',
    'VECM',
    'DynamicFactor',
    'WindowRegression',
    'VAR',
    'DatepartRegression',
    "GluonTS",
    "UnivariateRegression",
    'MultivariateRegression',
    'SectionalMotif',  # kinda
    'ARDL',
    'NeuralProphet',
    'ARCH',
    'Cassandra',
    'PreprocessingRegression',
]
motifs = [
    'UnivariateMotif',
    "MultivariateMotif",
    'SectionalMotif',
    'MotifSimulation',
    'MetricMotif',
    'SeasonalityMotif',
]
regressions = [
    'RollingRegression',
    'WindowRegression',
    'DatepartRegression',
    'UnivariateRegression',
    'MultivariateRegression',
    'PreprocessingRegression',
]
no_shared_fast = list(set(no_shared).intersection(set(fast_parallel_no_arima)))
# this should be implementable with some models in gluonts
all_result_path = [
    "UnivariateMotif",
    "MultivariateMotif",
    "SectionalMotif",
    'MetricMotif',
    "SeasonalityMotif",
    "Motif",
    "ARCH",  # simulations not motifs but similar
    "PytorchForecasting",
]
# these are those that require a parameter, and return a dict
diff_window_motif_list = [
    "UnivariateMotif",
    "MultivariateMotif",
    "Motif",
    "ARCH",
]
# models that fit and then have updated predicts without updated model fits (just data update)
update_fit = [
    'MultivariateRegression',
    "DatepartRegression",
    "GluonTS",
    'WindowRegression',
    'Cassandra',
    'PreprocessingRegression',
]
model_lists = {
    "all": all_models,
    "default": default,
    "fast": fast,
    "superfast": superfast,
    "parallel": parallel,
    "fast_parallel": fast_parallel,
    "fast_parallel_no_arima": fast_parallel_no_arima,
    "probabilistic": probabilistic,
    "multivariate": multivariate,
    "univariate": univariate,
    "no_params": no_params,
    "recombination_approved": recombination_approved,
    "no_shared": no_shared,
    "no_shared_fast": no_shared_fast,
    "experimental": experimental,
    "slow": slow,
    "gpu": gpu,
    "regressor": regressor,
    "best": best,
    "motifs": motifs,
    "all_result_path": all_result_path,
    "regressions": regressions,
    "all_pragmatic": all_pragmatic,
    "update_fit": update_fit,
}


def auto_model_list(n_jobs, n_series, frequency):
    pass


def model_list_to_dict(model_list):
    """Convert various possibilities to dict."""
    if model_list in list(model_lists.keys()):
        model_list = model_lists[model_list]

    if isinstance(model_list, dict):
        model_prob = list(model_list.values())
        model_list = [*model_list]
    elif isinstance(model_list, list):
        trs_len = len(model_list)
        model_prob = [1 / trs_len] * trs_len
    else:
        raise ValueError("model_list type not recognized.")
    return model_list, model_prob



# -*- coding: utf-8 -*-
"""Greykite."""
import datetime
import random
import pandas as pd
from autots.models.base import ModelObject, PredictionObject

try:
    from greykite.framework.templates.autogen.forecast_config import ForecastConfig
    from greykite.framework.templates.autogen.forecast_config import MetadataParam
    from greykite.framework.templates.forecaster import Forecaster
    from greykite.framework.templates.model_templates import ModelTemplateEnum
    from greykite.framework.templates.autogen.forecast_config import ComputationParam
    from greykite.algo.forecast.silverkite.constants.silverkite_holiday import (
        SilverkiteHoliday,
    )
    from greykite.framework.templates.autogen.forecast_config import (
        ModelComponentsParam,
    )
except Exception:
    _has_greykite = False
else:
    _has_greykite = True


def seek_the_oracle(
    df_index,
    series,
    col,
    forecast_length,
    freq,
    prediction_interval=0.9,
    model_template='silverkite',
    growth=None,
    holiday=True,
    holiday_country="UnitedStates",
    regressors=None,
    verbose=0,
    inner_n_jobs=1,
    **kwargs,
):
    """Internal. For loop or parallel version of Greykite."""
    inner_df = pd.DataFrame(
        {
            'ts': df_index,
            'y': series,
        }
    )
    if regressors is not None:
        inner_regr = regressors.copy()
        new_names = [
            'rrrr' + str(x) if x in inner_df.columns else str(x)
            for x in inner_regr.columns
        ]
        inner_regr.columns = new_names
        inner_regr.index.name = 'ts'
        inner_regr.reset_index(drop=False, inplace=True)
        inner_df = inner_df.merge(inner_regr, left_on='ts', right_on='ts', how='outer')
    metadata = MetadataParam(
        time_col="ts",  # name of the time column ("date" in example above)
        value_col="y",  # name of the value column ("sessions" in example above)
        freq=freq,  # "H" for hourly, "D" for daily, "W" for weekly, etc.
    )
    # INCLUDE forecast_length lagged mean and std of other features!
    model_template = ModelTemplateEnum.SILVERKITE.name
    forecaster = Forecaster()  # Creates forecasts and stores the result
    if regressors is not None:
        model_components = ModelComponentsParam(
            growth=growth, regressors={"regressor_cols": new_names}
        )
    else:
        model_components = ModelComponentsParam(
            growth=growth,  # 'linear', 'quadratic', 'sqrt'
        )
    computation = ComputationParam(n_jobs=inner_n_jobs, verbose=verbose)
    if holiday:  # also 'auto'
        model_components.events = {
            # These holidays as well as their pre/post dates are modeled as individual events.
            "holidays_to_model_separately": SilverkiteHoliday.ALL_HOLIDAYS_IN_COUNTRIES,  # all holidays in "holiday_lookup_countries"
            "holiday_lookup_countries": [
                holiday_country
            ],  # only look up holidays in the United States
            "holiday_pre_num_days": 1,  # also mark the 1 days before a holiday as holiday
            "holiday_post_num_days": 1,  # also mark the 1 days after a holiday as holiday
        }
    config = ForecastConfig(
        model_template=model_template,
        forecast_horizon=forecast_length,
        coverage=prediction_interval,
        model_components_param=model_components,
        metadata_param=metadata,
        computation_param=computation,
    )
    result = forecaster.run_forecast_config(  # result is also stored as `forecaster.forecast_result`.
        df=inner_df,
        config=config,
    )
    res_df = result.forecast.df.tail(forecast_length).drop(columns=['actual'])
    res_df['series_id'] = col
    return res_df


class Greykite(ModelObject):
    """Greykite

    Args:
        name (str): String to identify class
        frequency (str): String alias of datetime index frequency or else 'infer'
        prediction_interval (float): Confidence interval for probabilistic forecast
        holiday (bool): If true, include holidays
        regression_type (str): type of regression (None, 'User')

    """

    def __init__(
        self,
        name: str = "Greykite",
        frequency: str = 'infer',
        prediction_interval: float = 0.9,
        holiday: bool = False,
        growth: str = None,
        regression_type: str = None,
        holiday_country: str = 'US',
        random_seed: int = 2020,
        verbose: int = 0,
        n_jobs: int = None,
    ):
        ModelObject.__init__(
            self,
            name,
            frequency,
            prediction_interval,
            regression_type=regression_type,
            holiday_country=holiday_country,
            random_seed=random_seed,
            verbose=verbose,
            n_jobs=n_jobs,
        )
        self.holiday = holiday
        self.growth = growth

    def fit(self, df, future_regressor=[]):
        """Train algorithm given data supplied.

        Args:
            df (pandas.DataFrame): Datetime Indexed
        """
        if not _has_greykite:
            raise ImportError("Package greykite is required")

        df = self.basic_profile(df)
        self.regressor_train = None

        if self.regression_type == 'User':
            self.regressor_train = future_regressor.copy()

        self.df_train = df

        self.fit_runtime = datetime.datetime.now() - self.startTime
        return self

    def predict(
        self,
        forecast_length: int,
        future_regressor=[],
        just_point_forecast: bool = False,
    ):
        """Generates forecast data immediately following dates of index supplied to .fit()

        Args:
            forecast_length (int): Number of periods of data to forecast ahead
            regressor (numpy.Array): additional regressor, not used
            just_point_forecast (bool): If True, return a pandas.DataFrame of just point forecasts

        Returns:
            Either a PredictionObject of forecasts and metadata, or
            if just_point_forecast == True, a dataframe of point forecasts
        """
        if not _has_greykite:
            raise ImportError("Package greykite is required")
        predictStartTime = datetime.datetime.now()
        regressors = None
        if self.regression_type == 'User':
            regressors = pd.concat([self.regressor_train, future_regressor])

        parallel = True
        cols = self.df_train.columns.tolist()
        if self.n_jobs in [0, 1] or len(cols) < 4:
            parallel = False
        else:
            try:
                from joblib import Parallel, delayed
            except Exception:
                parallel = False
        # joblib multiprocessing to loop through series
        if parallel:
            verbs = 0 if self.verbose < 1 else self.verbose - 1
            df_list = Parallel(n_jobs=self.n_jobs, verbose=(verbs))(
                delayed(seek_the_oracle)(
                    self.df_train.index,
                    self.df_train[col],
                    col,
                    forecast_length,
                    freq=self.frequency,
                    prediction_interval=self.prediction_interval,
                    growth=dict(growth_term=self.growth),
                    holiday=self.holiday,
                    holiday_country=self.holiday_country,
                    regressors=regressors,
                    inner_n_jobs=self.n_jobs,
                )
                for col in cols
            )
            complete = pd.concat(df_list)
        else:
            df_list = []
            for col in cols:
                df_list.append(
                    seek_the_oracle(
                        self.df_train.index,
                        self.df_train[col],
                        col,
                        forecast_length,
                        freq=self.frequency,
                        prediction_interval=self.prediction_interval,
                        growth=dict(growth_term=self.growth),
                        holiday=self.holiday,
                        holiday_country=self.holiday_country,
                        regressors=regressors,
                        inner_n_jobs=self.n_jobs,
                    )
                )
            complete = pd.concat(df_list)

        forecast = complete.pivot_table(
            values="forecast", index="ts", columns="series_id", aggfunc="sum"
        )

        if just_point_forecast:
            return forecast
        else:
            upper_forecast = complete.pivot_table(
                values="forecast_upper", index="ts", columns="series_id", aggfunc="sum"
            )
            lower_forecast = complete.pivot_table(
                values="forecast_lower", index="ts", columns="series_id", aggfunc="sum"
            )

            predict_runtime = datetime.datetime.now() - predictStartTime
            prediction = PredictionObject(
                model_name=self.name,
                forecast_length=forecast_length,
                forecast_index=forecast.index,
                forecast_columns=forecast.columns,
                lower_forecast=lower_forecast,
                forecast=forecast,
                upper_forecast=upper_forecast,
                prediction_interval=self.prediction_interval,
                predict_runtime=predict_runtime,
                fit_runtime=self.fit_runtime,
                model_parameters=self.get_params(),
            )

            return prediction

    def get_new_params(self, method: str = 'random'):
        """Return dict of new parameters for parameter tuning."""
        holiday_choice = random.choices([True, False], [0.2, 0.8])[0]
        regression_list = [None, 'User']
        regression_probability = [0.8, 0.2]
        regression_choice = random.choices(regression_list, regression_probability)[0]
        growth_choice = random.choices(
            [None, 'linear', 'quadratic', 'sqrt'], [0.3, 0.3, 0.1, 0.1]
        )[0]

        parameter_dict = {
            'holiday': holiday_choice,
            'regression_type': regression_choice,
            'growth': growth_choice,
        }
        return parameter_dict

    def get_params(self):
        """Return dict of current parameters."""
        parameter_dict = {
            'holiday': self.holiday,
            'regression_type': self.regression_type,
            'growth': self.growth,
        }
        return parameter_dict



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------
from datetime import date
project = 'AutoTS'
copyright = u'%s, Colin Catlin' % date.today().year
author = 'Colin Catlin'

# The full version, including alpha/beta/rc tags
# import AutoTS
# from  AutoTS import __version__
# release = __version__
release = "0.6.1"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# Add napoleon to the extensions list
# 'recommonmark',
extensions = ['sphinx.ext.napoleon', 'sphinx.ext.autodoc', 'm2r2', 'sphinx.ext.githubpages',]

source_suffix = ['.rst', '.md']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

html_theme_options = {
    "show_powered_by": False,
	'analytics_id': 'G-P2KLF8302E',
	'logo': 'autots_logo.png',
	'description': 'Automated Forecasting',
    "github_user": "winedarksea",
    "github_repo": "autots",
    "github_banner": False,
    "show_related": False,
    "note_bg": "#FFF59C",
}
# Output file base name for HTML help builder.
htmlhelp_basename = "autotsdoc"


"""
Tools for Importing Sample Data
"""
from autots.datasets._base import (
    load_daily,
    load_live_daily,
    load_monthly,
    load_yearly,
    load_hourly,
    load_weekly,
    load_weekdays,
    load_zeroes,
    load_linear,
    load_sine,
    load_artificial,
)

__all__ = [
    'load_daily',
    'load_monthly',
    'load_yearly',
    'load_hourly',
    'load_weekly',
    'load_weekdays',
    'load_live_daily',
    'load_zeroes',
    'load_linear',
    'load_sine',
    'load_artificial',
]
